#!/usr/bin/env python

import neuroglancer
import numpy as np
import signal
import sys

def stop(signal, frame):
    neuroglancer.stop()
    sys.exit(0)

a = np.ones((100,100,100), dtype=np.uint8)*255
b = np.random.randint(0, 100, (100,1000,1000), dtype=np.uint64)

layers = [
    ('a', a),
    ('b', b)
]

print neuroglancer.serve(layers, server_args = { 'bind_address': '127.0.0.1', 'bind_port': 8484 })

signal.signal(signal.SIGINT, stop)
print('Server started, press Ctrl+C to stop')
signal.pause()



# @license
# Copyright 2016 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import collections
import json

from . import volume

try:
  # Python 2 case
  from urllib import quote as urlquote
except ImportError:
  # Python 3 case
  from urllib.parse import quote as urlquote


class Layer(object):

  def __init__(self,
               data,
               name=None,
               default_voxel_size=(1, 1, 1),
               voxel_size=None,
               offset=None,
               shader=None,
               visible=None,
               **kwargs):
    if offset is None:
      offset = (0, 0, 0)
      if hasattr(data, 'attrs'):
        if 'resolution' in data.attrs:
          voxel_size = tuple(data.attrs['resolution'])[::-1]
        if 'offset' in data.attrs:
          offset = tuple(data.attrs['offset'])[::-1]
    if voxel_size is None:
      voxel_size = default_voxel_size
    self.volume = volume.ServedVolume(data=data,
                                      offset=offset,
                                      voxel_size=voxel_size,
                                      **kwargs)
    self.name = name
    extra_args = self.extra_args = dict()
    if shader is not None:
      extra_args['shader'] = shader
    if visible is not None:
      extra_args['visible'] = visible

  def get_layer_spec(self, server_url):
    return dict(type=self.volume.volume_type,
                source='python://%s/%s' % (server_url, self.volume.token),
                **self.extra_args)


class BaseViewer(object):

  def __init__(self, voxel_size=(1, 1, 1)):
    self.voxel_size = voxel_size
    self.layers = []

  def add(self, *args, **kwargs):
    layer = Layer(*args, default_voxel_size=self.voxel_size, **kwargs)
    self.layers.append(layer)

  def get_json_state(self):
    state = collections.OrderedDict()
    layers = state['layers'] = collections.OrderedDict()
    specified_names = set(layer.name for layer in self.layers)
    for layer_i, layer in enumerate(self.layers):
      self.register_volume(layer.volume)
      name = layer.name
      if name is None:
        base_name = layer.volume.volume_type
        name = base_name
        suffix = 2
        while name in specified_names:
          name = '%s%d' % (base_name, suffix)
          suffix += 1
        specified_names.add(name)
      layers[name] = layer.get_layer_spec(self.get_server_url())
    return state

  def register_volume(self, volume):
    """Registers a volume with the server.

    This must be overridden by a subclass.
    """
    raise NotImplemented

  def get_server_url(self):
    """Returns the root URL for the server, e.g. 'http://hostname:port'.

    This must be overridden by a subclass.
    """
    raise NotImplemented

  def get_encoded_state(self):
    return urlquote(
        json.dumps(self.get_json_state(),
                   separators=(',', ':')),
        '~@#$&()*!+=:;,.?/\'')



# @license
# Copyright 2016 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pkg_resources
import posixpath
import os

content = {
    k: (pkg_resources.resource_string(__name__, k) if
        pkg_resources.resource_exists(__name__, k) else None)
    for k in ['main.bundle.js', 'chunk_worker.bundle.js', 'styles.css',
              'index.html']
}

main_js = content['main.bundle.js']
chunk_worker_js = content['chunk_worker.bundle.js']
styles_css = content['styles.css']
index_html = content['index.html']

mime_type_map = {
    '.css': 'text/css',
    '.js': 'application/javascript',
    '.html': 'text/html',
    '.map': 'application/json'
}

def guess_mime_type_from_path(path):
  return mime_type_map.get(posixpath.splitext(path)[1],
                           'application/octet-stream')

class StaticContentSource(object):
  def get(self, name):
    if name == '':
      name = 'index.html'
    return self.get_content(name), guess_mime_type_from_path(name)
  def get_content(self, name):
    raise NotImplemented

class PkgResourcesContentSource(StaticContentSource):
  def get_content(self, name):
    if name not in content:
      raise ValueError('Invalid static resource name: %r' % name)
    value = content.get(name)
    if value is None:
      raise ValueError(
          'Static resources not built.  Run: "python setup.py bundle_client" or use an alternative static content source.')
    return value

class HttpSource(StaticContentSource):
  def __init__(self, url):
    self.url = url

  def get_content(self, name):
    import requests
    full_url = posixpath.join(self.url, name)
    r = requests.get(full_url)
    if r.status_code >= 200 and r.status_code < 300:
      return r.content
    raise ValueError('Failed to retrieve %r: %s' % (full_url, r.reason))

class FileSource(StaticContentSource):
  def __init__(self, path=None, open=open):
    if path is None:
      path = os.path.join(os.path.dirname(__file__), '../../../dist/dev')
    self.file_path = path
    self.file_open = open

  def get_content(self, name):
    full_path = os.path.join(self.file_path, name)
    try:
      with self.file_open(full_path, 'rb') as f:
        return f.read()
    except Exception as e:
      raise ValueError('Failed to read local path %r: %s' % (full_path, e))

dist_dev_static_content_source = FileSource()

def get_default_static_content_source():
  return PkgResourcesContentSource()

def get_static_content_source(
    source=None,
    url=None,
    path=None,
    open=None):
  if source is not None:
    return source
  elif url is not None:
    return HttpSource(url)
  elif path is not None:
    return FileSource(path=path, open=open)
  else:
    return get_default_static_content_sourceb()



# @license
# Copyright 2016 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from . import base_viewer
from . import server

class Viewer(base_viewer.BaseViewer):
  def __init__(self, *args, **kwargs):
    super(Viewer, self).__init__(*args, **kwargs)
    server.start()

  def register_volume(self, volume):
    server.register_volume(volume)

  def get_server_url(self):
    return server.get_server_url()

  def get_viewer_url(self):
    return '%s/static/%s/#!%s' % (self.get_server_url(),
                                  server.global_server.token,
                                  self.get_encoded_state())

  def __repr__(self):
    return self.get_viewer_url()

  def _repr_html_(self):
    return '<a href="%s" target="_blank">Viewer</a>' % self.get_viewer_url()

def view(*args, **kwargs):
  v = Viewer()
  v.add(*args, **kwargs)
  return v



# @license
# Copyright 2016 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from PIL import Image
import numpy as np
import zlib
import io

def encode_jpeg(subvol):
  shape = subvol.shape
  reshaped = subvol.reshape(shape[0] * shape[1], shape[2])
  img = Image.fromarray(reshaped)
  f = io.BytesIO()
  img.save(f, "JPEG")
  return f.getvalue()
def encode_npz(subvol):
  fileobj = io.BytesIO()
  if len(subvol.shape) == 3:
    subvol = np.expand_dims(subvol, 0)
  np.save(fileobj, subvol)
  cdz = zlib.compress(fileobj.getvalue())
  return cdz
def encode_raw(subvol):
  return subvol.tostring('C')



# @license
# Copyright 2016 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, print_function

try:
  # Python 2 case
  from SocketServer import ThreadingMixIn
  from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler
except ImportError:
  # Python 3 case
  from socketserver import ThreadingMixIn
  from http.server import HTTPServer, BaseHTTPRequestHandler

import threading
import re
import os
import json
import posixpath
import socket
from .volume import ServedVolume
from .token import make_random_token
from . import static

info_path_regex = re.compile(r'^/neuroglancer/info/([^/]+)$')

data_path_regex = re.compile(
    r'^/neuroglancer/([^/]+)/([^/]+)/([0-9]+),([0-9]+)/([0-9]+),([0-9]+)/([0-9]+),([0-9]+)$')

mesh_path_regex = re.compile(
    r'^/neuroglancer/mesh/([^/]+)/([0-9]+)$')

static_path_regex = re.compile(
    r'/static/([^/]+)/((?:[a-zA-Z0-9_\-][a-zA-Z0-9_\-.]*)?)$')

global_static_content_source = None

global_server_args = dict(
    bind_address='127.0.0.1',
    bind_port=0)

class Server(ThreadingMixIn, HTTPServer):
  def __init__(self,
               bind_address='127.0.0.1',
               bind_port=0):
    HTTPServer.__init__(self, (bind_address, bind_port), RequestHandler)
    self.daemon_threads = True
    self.volumes = dict()
    self.token = make_random_token()
    global global_static_content_source
    if global_static_content_source is None:
      global_static_content_source = static.get_default_static_content_source()

    if bind_address == '0.0.0.0':
      hostname = socket.getfqdn()
    else:
      hostname = bind_address
    self.server_url = 'http://%s:%s' % (hostname, self.server_address[1])

class RequestHandler(BaseHTTPRequestHandler):
  protocol_version = 'HTTP/1.1'
  def do_GET(self):
    m = re.match(data_path_regex, self.path)
    if m is not None:
      self.handle_data_request(
          token=m.group(2),
          data_format=m.group(1),
          start=(int(m.group(3)), int(m.group(5)), int(m.group(7))),
          end=(int(m.group(4)), int(m.group(6)), int(m.group(8))))
      return
    m = re.match(mesh_path_regex, self.path)
    if m is not None:
      self.handle_mesh_request(m.group(1), int(m.group(2)))
      return
    m = re.match(info_path_regex, self.path)
    if m is not None:
      self.handle_info_request(m.group(1))
      return
    m = re.match(static_path_regex, self.path)
    if m is not None:
      self.handle_static_request(m.group(1), m.group(2))
      return
    self.send_error(404)

  def handle_static_request(self, token, path):
    if token != self.server.token:
      self.send_error(404)
    try:
      data, content_type = global_static_content_source.get(path)
    except ValueError as e:
      self.send_error(404, e.args[0])
      return

    self.send_response(200)
    self.send_header('Content-type', content_type)
    self.send_header('Content-length', len(data))
    self.end_headers()
    self.wfile.write(data)

  def handle_data_request(self, token, data_format, start, end):
    volume = self.server.volumes.get(token)
    if volume is None:
      self.send_error(404)
      return
    try:
      data, content_type = volume.get_encoded_subvolume(data_format, start, end)
    except ValueError as e:
      self.send_error(400, x.args[0])
      return

    self.send_response(200)
    self.send_header('Content-type', content_type)
    self.send_header('Content-length', len(data))
    self.send_header('Access-Control-Allow-Origin', '*')
    self.end_headers()
    self.wfile.write(data)

  def handle_info_request(self, token):
    volume = self.server.volumes.get(token)
    if volume is None:
      self.send_error(404)
      return
    data = json.dumps(volume.info())
    self.send_response(200)
    self.send_header('Content-type', 'application/json')
    self.send_header('Content-length', len(data))
    self.send_header('Access-Control-Allow-Origin', '*')
    self.end_headers()
    self.wfile.write(data)

  def handle_mesh_request(self, key, object_id):
    volume = self.server.volumes.get(key)
    if volume is None:
      self.send_error(404)
    try:
      encoded_mesh = volume.get_object_mesh(object_id)
    except ValueError as e:
      self.send_error(400, e.args[0])
      return
    self.send_response(200)
    self.send_header('Content-type', 'application/octet-stream')
    self.send_header('Content-length', len(data))
    self.send_header('Access-Control-Allow-Origin', '*')
    self.end_headers()
    self.wfile.write(data)

global_server = None

def set_static_content_source(*args, **kwargs):
  global global_static_content_source
  global_static_content_source = static.get_static_content_source(*args, **kwargs)

def set_server_bind_address(bind_address='127.0.0.1', bind_port=0):
  global_server_args = dict(bind_address=bind_address, bind_port=bind_port)

def is_server_running():
  return global_server is not None

def stop():
  global global_server
  if global_server is not None:
    global_server.shutdown()
    global_server = None

def get_server_url():
  return global_server.server_url

def start():
  global global_server
  if global_server is None:
    global_server = Server(**global_server_args)
    thread = threading.Thread(target=lambda: global_server.serve_forever())
    thread.daemon = True
    thread.start()

def register_volume(volume):
  start()
  global_server.volumes[volume.token] = volume



#!/usr/bin/env python

from setuptools import setup, find_packages
from distutils.command.build import build
from subprocess import call
import os
import shutil

static_files = [ 'main.bundle.js', 'chunk_worker.bundle.js', 'styles.css', 'index.html' ]

class bundle_client(build):

    user_options = [

        ('client-bundle-type=', None, 'The nodejs bundle type. "min" (default) creates condensed static files for production, "dev" creates human-readable files.')
    ]

    def initialize_options(self):

        self.client_bundle_type = 'min'

    def finalize_options(self):

        if self.client_bundle_type not in  [ 'min', 'dev' ]:
            raise RuntimeError('client-bundle-type has to be one of "min" or "dev"')

    def run(self):

        this_dir = os.path.abspath(os.path.dirname(__file__))
        project_dir = os.path.join(this_dir, '..')

        build_dir = os.path.join(project_dir, 'dist/' + self.client_bundle_type)
        static_dir = os.path.join(this_dir, 'neuroglancer/static')

        print "Project dir " + project_dir
        print "Build dir " + build_dir
        print "Static dir " + static_dir

        prev_dir = os.path.abspath('.')
        os.chdir(project_dir)

        target = {
            "min" : "build-min",
            "dev" : "build"
        }

        try:
            call(['npm', 'i'])
            res = call(['npm', 'run', target[self.client_bundle_type]])
        except:
            raise RuntimeError('Could not run \'npm run build-%s\'. Make sure node.js >= v5.9.0 is installed and in your path.' % self.client_bundle_type)

        if res != 0:
            raise RuntimeError('failed to bundle neuroglancer node.js project')

        try:
            os.mkdir(static_dir)
        except OSError:
            pass

        for f in static_files:
            shutil.copy(
                os.path.join(build_dir, f),
                os.path.join(static_dir, f)
            )

        os.chdir(prev_dir)

setup(
    name = 'neuroglancer',
    version = '0.0.6',
    description = 'Python data backend for neuroglancer, a WebGL-based viewer for volumetric data',
    author = 'Jeremy Maitin-Shepard, Jan Funke',
    author_email = 'jbms@google.com, jfunke@iri.upc.edu',
    url = 'https://github.com/google/neuroglancer',
    license = 'Apache License 2.0',
    packages = find_packages(),
    package_data = {
        'neuroglancer.static': static_files,
    },
    install_requires = [
        "Pillow>=3.2.0",
        "numpy>=1.11.0",
    ],
    use_2to3 = True,
    cmdclass = {'bundle_client' : bundle_client},
)



# @license
# Copyright 2016 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import numpy as np

from .token import make_random_token
from .chunks import encode_jpeg, encode_npz, encode_raw

class ServedVolume(object):
  def __init__(self,
               data,
               offset=(0, 0, 0),
               voxel_size=(1, 1, 1),
               encoding='npz',
               chunk_data_sizes=None,
               volume_type=None):
    """Initializes a ServedVolume.

    @param data 3-d [z, y, x] array or 4-d [channel, z, y, x] array.
    """
    self.token = make_random_token()
    if len(data.shape) == 3:
      self.num_channels = 1
      self.shape = data.shape[::-1]
    else:
      if len(data.shape) != 4:
        raise ValueError('data array must be 3- or 4-dimensional.')
      self.num_channels = data.shape[0]
      self.shape = data.shape[1:][::-1]

    self.data = data
    self.voxel_size = voxel_size
    self.offset = offset
    self.data_type = data.dtype.name
    self.encoding = encoding
    if chunk_data_sizes is not None:
      arr = np.array(chunk_data_sizes)
      if (len(arr.shape) != 2 or arr.shape[1] != 3 or np.any(arr < 1) or
          np.any(np.cast[int](arr) != arr)):
        raise ValueError(
            'chunk_data_sizes must be a sequence of 3-element non-negative integers')
    self.chunk_data_sizes = chunk_data_sizes
    if volume_type is None:
      if self.num_channels == 1 and (self.data_type == 'uint16' or
                                     self.data_type == 'uint32' or
                                     self.data_type == 'uint64'):
        volume_type = 'segmentation'
      else:
        volume_type = 'image'
    self.volume_type = volume_type

  def info(self):
    upper_voxel_bound = tuple(np.array(self.offset) + np.array(self.shape))
    info = dict(volumeType=self.volume_type,
                dataType=self.data_type,
                encoding=self.encoding,
                numChannels=self.num_channels,
                scales=[
                    dict(key=self.token,
                         lowerVoxelBound=self.offset,
                         upperVoxelBound=upper_voxel_bound,
                         voxelSize=self.voxel_size),
                ])
    if self.chunk_data_sizes is not None:
      info['chunkDataSizes'] = self.chunk_data_sizes
    return info

  def get_encoded_subvolume(self, data_format, start, end):
    offset = self.offset
    shape = self.shape
    for i in xrange(3):
      if end[i] < start[i] or offset[i] > start[i] or end[i] - offset[i] > shape[i]:
        raise ValueError('Out of bounds data request.')

    indexing_expr = tuple(np.s_[start[i] - offset[i]:end[i] - offset[i]] for i in (2,1,0))
    if len(self.data.shape) == 3:
      subvol = self.data[indexing_expr]
    else:
      subvol = self.data[(np.s_[:],) + indexing_expr]
    content_type = 'application/octet-stream'
    if data_format == 'jpeg':
      data = encode_jpeg(subvol)
      content_type = 'image/jpeg'
    elif data_format == 'npz':
      data = encode_npz(subvol)
    elif data_format == 'raw':
      data = encode_raw(subvol)
    else:
      raise ValueError('Invalid data format requested.')
    return data, content_type

  def get_object_mesh(self, object_id):
    raise ValueError('Meshes not yet supported.')
